"""Dispatcher is the class responsible for calling the correct overloaded
function based on received arguments. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Callable, Any

Overloaded = dict[tuple[type, ...], Callable]


class Dispatcher:
  """Dispatcher is the class responsible for calling the correct overloaded
  function based on received arguments. """

  __overloaded_functions__ = None

  def __init__(self, overloadedFunctions: Overloaded) -> None:
    self.__overloaded_functions__ = overloadedFunctions

  def functionFactory(self, ) -> Callable:
    """Return a function that calls the dispatcher."""

    def callMeMaybe(this: Any, *args) -> Any:
      """Call the dispatcher."""
      typeSig = (*[type(arg) for arg in args],)
      func = self.__overloaded_functions__.get(typeSig, None)
      if func is None:
        args = [float(arg) if isinstance(arg, int) else arg for arg in args]
        typeSig = (*[type(arg) for arg in args],)
        func = self.__overloaded_functions__.get(typeSig, None)
      if callable(func):
        return self.__overloaded_functions__.get(typeSig, None)(this, *args)
      print(typeSig)
      return NotImplemented

    return callMeMaybe

  def staticFactory(self, ) -> Callable:
    """Return a static method that calls the dispatcher."""

    def callMeMaybe(*args) -> Any:
      """Call the dispatcher."""
      typeSig = tuple([type(arg) for arg in args])
      return self.__overloaded_functions__.get(typeSig, None)(*args)

    return callMeMaybe



"""BaseMetaclass provides general functionality for derived classes. This
includes primarily function overloading. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.meta import AbstractMetaclass, Bases, BaseNamespace


class BaseMetaclass(AbstractMetaclass):
  """BaseMetaclass provides general functionality for derived classes. This
  includes primarily function overloading. """

  @classmethod
  def __prepare__(mcls, name: str, bases: Bases, **kwargs) -> BaseNamespace:
    """The __prepare__ method is invoked before the class is created. This
    implementation ensures that the created class has access to the safe
    __init__ and __init_subclass__ through the BaseObject class in its
    method resolution order."""
    return BaseNamespace(mcls, name, bases, **kwargs)

  def __new__(mcls,
              name: str,
              bases: Bases,
              space: BaseNamespace,
              **kwargs) -> type:
    """The __new__ method is invoked to create the class."""
    namespace = space.compile()
    return super().__new__(mcls, name, bases, namespace, **kwargs)
